% This program produces Monte Carlo results for the semiparametrically efficient estimator
% applied to Example 1.
% This program requires two free Matlab toolboxes: YALMIP and MPT.
% Please set path to the directories where YALMIP and MPT are saved.
% This code uses Matlab's parallel processing toolbox.
% If prallel processing is not used, comment out line 19 and
% replace "parfor" in line 95 with "for".
% The user must set K (# of support points), c (strength of identification)
% and SI (# of inner loops).
% The program conducts S=5000 Monte Carlo simulations, which are split 
% into SI inner loops to save memory used by YALMIP.
% Pick SI so that S/SI becomes an integer, ideally a multiple of the number
% of processors.

clear
close all
clc
yalmip('clear') % clear memory used by yalmip
matlabpool(4)   % set # of processors for parallel processing

% Set simulations parameters %
n = 1000; 	% sample size
S = 5000;   % # of simulation draws
SI= 40;  	% # of inner loops (make sure S/SI is an integer)
B = 200; 	% # of booststrap samples
m = 100;  	% # of grid points on unit sphere
K = 5;   	% # of support points
c = 1;  	% strength of identification
suppz = linspace(-5,5,K); %z's support
tau   = [0, log(log(n)), log(n), n^(1/8), n^(1/4)]; % grid points for slackness
A = [-ones(K,1) -suppz'; ones(K,1) suppz']; % slope of affine constraints

% Set parameter values (DGP)
beta  = [1; 2];     % true parameter
dt = size(beta,1);	% dimension of parameter
dF = 2*K; 			% number of inequalities
dtau = length(tau); % number of values of tau
dloc = 5;			% number of local alternatives

% Initialize cells & matrices that store results %
SupportFunction = zeros(m,S);                    % a matrix that stores support function of KS's estimate w/o correction
LMultiplier = cell(S,1);                         % a cell that stores simulated LMs w/o correction
StKS_kink = zeros(S,dloc);                       % a matrix that stores KS test stat for a kink point
StKS_flat = zeros(S,dloc);                       % a matrix that stores KS test stat for a point on a flat face
C1 = zeros(S,1);                                 % a vector that stores expansion factor for CS1 w/o correction
CKS_kink = zeros(S,dtau,dloc);                   % a vector that stores KS critical values for a kink point
CKS_flat = zeros(S,dtau,dloc);                   % a vector that stores KS critical values for point on a flat face
VolKS    = zeros(S,1);                           % a vector that stores KS's CI's volume w/o correction

% Stopwatch %
t_prim = zeros(S,1);			% time needed to solve primal problem
t_dual = zeros(S,1);            % time needed to solve dual problem
t_eval = zeros(S,1);            % time needed to evaluate value function
t_weig = zeros(S,1);            % time needed to generate random weights
t_boot = zeros(S,1);            % time needed to evalate bootstrapped process
t_crit = zeros(S,1);            % time needed to compute bootstrap critical value for CS 
t_kink = zeros(S,dtau);			% time needed to compute bootstrap critical value for kink 
t_flat = zeros(S,dtau);			% time needed to compute bootstrap critical value for flat face 

% Temporary vectors for inner loops %
subSupportFunction = zeros(m,S/SI); 
subLMultiplier = cell(S/SI,1);
subStKS_kink = zeros(S/SI,dloc);
subStKS_flat = zeros(S/SI,dloc);
subC1 = zeros(S/SI,1);
subCKS_kink = zeros(S/SI,dtau,dloc);
subCKS_flat = zeros(S/SI,dtau,dloc);
subvolKS    = zeros(S/SI,1);
subt_prim = zeros(S/SI,1);
subt_dual = zeros(S/SI,1);
subt_eval = zeros(S/SI,1);
subt_weig = zeros(S/SI,1);
subt_boot = zeros(S/SI,1);
subt_kink = zeros(S/SI,dtau);
subt_flat = zeros(S/SI,dtau);

% Grid the unit sphere %
th = linspace(-pi,pi,m); % angle
q = [sin(th);cos(th)]';  % unit sphere

% Load Population Data %
load(['./mat/population_K' num2str(K) 'c' num2str(c)  '.mat'])
loc_kink = repmat(theta_kink',1,dloc) + repmat([0;1],1,dloc) .* repmat(linspace(0,10,dloc)./sqrt(n),2,1); % kink point and alternatives
loc_flat = repmat(theta_flat',1,dloc) + repmat([1;0],1,dloc) .* repmat(linspace(0,10,dloc)./sqrt(n),2,1); % point on flat face and alternatives

ops  = sdpsettings('cachesolvers',1);
% The beginning of Monte Carlo %

for h=1:SI
    % Define variables used by YALMIP outside the inner loop %
    thetas = sdpvar(dt,1); 	% control variable in the primal problem
    w = sdpvar(dF,1); 		% control variable in the dual problem
    p = sdpvar(dt,1); 		% unit vector
    % The beginning of inner loop using parallel processing
    parfor s=1:S/SI
        si = s+(h-1)*(S/SI);
        stream = RandStream.create('mlfg6331_64','NumStreams',S,'StreamIndices',1:S,'seed',0,'CellOutput',1); 
        RandStream.setDefaultStream(stream{si});
        % Data Generation %
        Z = [ones(n,1) randsample(stream{si},suppz,n,true)']; 	 % generate Z
        e = randn(stream{si},n,1);     							 % generate error for Y
        v = 0.2*rand(stream{si},n,1); 							 % generate random coefficients
        ticweig = tic;
        W = reshape(randsample(stream{si},[-1,1],n*B,true),n,B); % generate bootstrap weights
        subt_weig(s) = toc(ticweig);
        
        % Generate YLi, YUi
        Y  = Z*beta + e;          % generate Y
        YL = Y -v.*Z(:,2).^2 -c ; % generate Y_L
        YU = Y +v.*Z(:,2).^2 +c ; % generate Y_U
        
        % Instrument (indiators) %
        Iz = repmat(Z(:,2),1,length(suppz)) == repmat(suppz,n,1);    	% indicator for each value of z                      
        EYLz = Iz'*YL./n;   											% mean of Y_L * Iz							
        EYUz = Iz'*YU./n;   											% mean of Y_U * Iz
        EIz = mean(Iz)';    											% mean of Iz
        
        % Variance of moments
        sYLz = mean(((Iz.*repmat(YL,1,K))./(repmat(EIz',n,1))-Iz.*repmat((EYLz./EIz)',n,1)).^2)'; 	% variance of moments associated with YL
        sYUz = mean(((Iz.*repmat(YU,1,K))./(repmat(EIz',n,1))-Iz.*repmat((EYUz./EIz)',n,1)).^2)'; 	% variance of moments associated with YU                                
        
        % Elements of Constraints %
        FL = [-EYLz./EIz; EYUz./EIz];                % intercept of affine constraints
        
         % Estimation (Kaido and Santos) %
        nonempt_est =0;                              % indicator for non-empty estimator             						
            
        % Define and solve the Primal Problem %
        tic
        Fp   = [A*thetas - FL <=  0, -1<=p<=1]; 	 % constraint in primal form
        objp = -p'*thetas;                     		 % objective function in primal form
        [~, ~, ~, Valuefunction_p, Optimizer_p] ...  % solve the primal problem
            = solvemp(Fp,objp,ops,p);
        subt_prim(s) = toc;
        
        % Define and solve the Dual Problem %
        tic
        Fd   = [A'*w == p, w>=0, -1<=p<=1];    		 % constraint in dual form
        objd = w'*FL;                                % objective function in dual form
        [~, ~, ~, ~,Optimizer_d]...
            = solvemp(Fd,objd,ops,p); 				 % solve the dual problem
        subt_dual(s) = toc;
        
        % Evaluate the support function and Lagrange multipliers %
        tic
        lambda = NaN(m,dF);                          % a matrix that stores LM
        theta = NaN(m,dt);                           % a matrix that stores the boundary points (from primal)
        mu = zeros(m,1);                             % a vector that stores support func
        for k = 1:m
            assign(p,q(k,:)');                       % set direction
            temp = double(Optimizer_p);              % store boundary point (KS: primal)            
            if ~isempty(temp)                        % if estimate is non-empty
                theta(k,:) = temp';                  % store the boundary of estimate
                mu(k) = - double(Valuefunction_p);   % store support function at a point
                lambda(k,:) = double(Optimizer_d)';  % store lagrange multiplier
            end
        end
        
        % Store how long it took and support function
        subt_eval(s) = toc;               
        subSupportFunction(:,s) = mu;            
        if ~(isnan(sum(theta(:,1))))                 % nonempty estimate found, exit            
            subLMultiplier{s} = lambda;              % store the whole LM function
            nonempt_est = 1;
            
            % Compute test statistic under null and alternative %
            subStKS_kink(s,:) = sqrt(n)*max(max(q*loc_kink-repmat(mu,1,dloc),0)); % test stat for point on kink
            subStKS_flat(s,:) = sqrt(n)*max(max(q*loc_flat-repmat(mu,1,dloc),0)); % test stat for point on flat face
        end
        
        % Multiplier Bootstrap %
        if nonempt_est==1
            ticboot = tic;
            mL = [Iz.*repmat(YL,1,K),Iz.*repmat(YU,1,K),Iz];            % compute mL
            nablaF = [-diag(1./EIz),zeros(K,K),diag(EYLz./((EIz).^2))   % compute nablaF
                zeros(K,K),diag(1./EIz),-diag(EYUz./((EIz).^2))];
            Ewm = (mL-repmat(mean(mL),n,1))'*W/n;          				% recenter mL and take weighted average
            GB = sqrt(n)*( - lambda * nablaF * Ewm);        		    % perturb g(p)'m(x)
            subt_boot(s) = toc(ticboot);
            ticstat_KS = tic;
            c1B = quantile(max(max(-GB,0)),0.95);           		    % expansion factor for CS
            subC1(s) = c1B;										        % store expansion factor for CS
        else
            ticstat_KS = tic;
            subC1(s) = 0;
            subt_boot(s) = 0;
        end
        subt_crit(s) = toc(ticstat_KS) + subt_boot(s);
        
        % Compute Volume of CS %
        if nonempt_est == 1
            theta_KS = theta + c1B.*q./sqrt(n);                     % compute 1-sided CS
            temp1 = unique(theta_KS,'rows');                		% extract unique elements of boundary
            try
                [~,vol_KS] = convhull(temp1(:,1),temp1(:,2));       % compute volume of CS
            catch exception
                vol_KS = 0;											% set volume to 0
            end
            subvolKS(s) = vol_KS;                                   % store volume of KS
        else														% if estimator is empty
            subvolKS(s) = 0;
        end
        
        % Compute Critical Value for Hypothesis Testing %
		Mn_kink = cell(dtau,dloc);  % set of directions for theta_kink
		Mn_flat = cell(dtau,dloc);  % set of directions for theta_hat
		c_kink  = zeros(dtau,dloc); % critical value for the test
		c_flat  = zeros(dtau,dloc); % critical value for the test
		t_kink_KS  = zeros(dtau,dloc);
		t_flat_KS  = zeros(dtau,dloc);
        for l=1:dtau
            for vv= 1:dloc
                if nonempt_est==1
                    tic
                    Mn_kink{l,vv} = find(q*loc_kink(:,vv)-mu >= max(q*loc_kink(:,vv)-mu)-tau(l)/sqrt(n));   % set (row indices) of directions for kink
                    if size(Mn_kink{l,vv},1)==1
                        c_kink(l,vv) = quantile(max(-GB(Mn_kink{l,vv},:),0),0.95);                          % compute critical value for kink
                    else
                        c_kink(l,vv) = quantile(max(max(-GB(Mn_kink{l,vv},:),0)),0.95); 					% compute critical value for kink
                    end
                    t_kink_KS(l,vv) = toc;
                    tic
                    Mn_flat{l,vv} = find(q*loc_flat(:,vv)-mu >= max(q*loc_flat(:,vv)-mu)-tau(l)/sqrt(n));   % set (row indices) of directions for flat face
                    if size(Mn_flat{l,vv},1)==1
                        c_flat(l,vv) = quantile(max(-GB(Mn_flat{l,vv},:),0),0.95);                          % compute critical value for flat face
                    else
                        c_flat(l,vv) = quantile(max(max(-GB(Mn_flat{l,vv},:),0)),0.95);                     % compute critical value for flat face
                    end
                    t_flat_KS(l,vv) = toc;
                end
            end
        end
        subCKS_kink(s,:,:) = c_kink;
        subCKS_flat(s,:,:) = c_flat;        
		subt_kink(s,:) = repmat(subt_boot(s),1,dtau) + sum(t_kink_KS');
		subt_flat(s,:) = repmat(subt_boot(s),1,dtau) + sum(t_flat_KS');
                 
    end
    
    % Pass results from the inner loop %
    SupportFunction(:,(h-1)*(S/SI)+1:h*(S/SI)) = subSupportFunction;
    LMultiplier((h-1)*(S/SI)+1:h*(S/SI))= subLMultiplier;
    StKS_kink((h-1)*(S/SI)+1:h*(S/SI),:) = subStKS_kink; 
    StKS_flat((h-1)*(S/SI)+1:h*(S/SI),:) = subStKS_flat;
    C1((h-1)*(S/SI)+1:h*(S/SI)) = subC1;
    CKS_kink((h-1)*(S/SI)+1:h*(S/SI),:,:) = subCKS_kink;
    CKS_flat((h-1)*(S/SI)+1:h*(S/SI),:,:) = subCKS_flat;
	VolKS((h-1)*(S/SI)+1:h*(S/SI))      = subvolKS;
    t_prim((h-1)*(S/SI)+1:h*(S/SI)) = subt_prim;
    t_dual((h-1)*(S/SI)+1:h*(S/SI)) = subt_dual;
    t_eval((h-1)*(S/SI)+1:h*(S/SI)) = subt_eval;
    t_weig((h-1)*(S/SI)+1:h*(S/SI)) = subt_weig;
    t_boot((h-1)*(S/SI)+1:h*(S/SI)) = subt_boot;
    t_crit((h-1)*(S/SI)+1:h*(S/SI)) = subt_crit;
    t_kink((h-1)*(S/SI)+1:h*(S/SI),:) = subt_kink;
    t_flat((h-1)*(S/SI)+1:h*(S/SI),:) = subt_flat;
                
    yalmip('clear') % Clear the memory used by YALMIP
    % Go to the next inner loop %
    disp(['Current sub-block: ' num2str(h) '/' num2str(SI)])
end
% The end of outer loop (Monte Carlo) %

% Check empty set estimates %
NEmpty_KS    = find(sum(abs(SupportFunction))~=0);	% find simulations with non-empty estimate
SKS    = size(NEmpty_KS,2);    						% calculate # of simulation draws with non-empty estimate

% Compute size and power %
spKS_kink = zeros(dtau,dloc);
spKS_flat = zeros(dtau,dloc);
spKS_kink_ne = zeros(dtau,dloc);
spKS_flat_ne = zeros(dtau,dloc);
for vv=1:dloc
spKS_kink(:,vv) = (sum(repmat(StKS_kink(NEmpty_KS,vv),1,dtau)-CKS_kink(NEmpty_KS,:,vv)>0)'+(S-SKS))/S; % compute size & power when theta is on kink
spKS_flat(:,vv) = (sum(repmat(StKS_flat(NEmpty_KS,vv),1,dtau)-CKS_flat(NEmpty_KS,:,vv)>0)'+(S-SKS))/S; % compute size & power when theta is on flat face
spKS_kink_ne(:,vv) = sum(repmat(StKS_kink(NEmpty_KS,vv),1,dtau)-CKS_kink(NEmpty_KS,:,vv)>0)'/SKS; % compute size & power (kink) cond. on initial estimate being nonempty
spKS_flat_ne(:,vv) = sum(repmat(StKS_flat(NEmpty_KS,vv),1,dtau)-CKS_flat(NEmpty_KS,:,vv)>0)'/SKS; % compute size & power (flat face) cond. on initial estimate being nonempty
end
spKS_kink = spKS_kink'; % each row gives size & power for each value of tau (kink)
spKS_flat = spKS_flat'; % each row gives size & power for each value of tau (flat face) 
spKS_kink_ne = spKS_kink_ne'; % each row gives size & power for each value of tau (kink) cond. on initial estimate being nonempty
spKS_flat_ne = spKS_flat_ne'; % each row gives size & power for each value of tau (flat face) cond. on initial estimate being nonempty

% Hausdorff distance loss for KS estimate %
Hausdorff = zeros(S,1);
InnerHausdorff = zeros(S,1);
OuterHausdorff = zeros(S,1);
Hausdorff(NEmpty_KS) = max(abs(SupportFunction(:,NEmpty_KS)-repmat(nuP0,1,SKS)));        % compute Hausdorff distance
InnerHausdorff(NEmpty_KS) = max(max(SupportFunction(:,NEmpty_KS)-repmat(nuP0,1,SKS),0)); % compute inner Hausdorff distance
OuterHausdorff(NEmpty_KS) = max(max(repmat(nuP0,1,SKS)-SupportFunction(:,NEmpty_KS),0)); % compute outer Hausdorff distance
Hausdorff(setdiff(1:S,NEmpty_KS)) = Inf;        % set Hausdorff loss = Inf if initial estimate is empty
InnerHausdorff(setdiff(1:S,NEmpty_KS)) = 0;     % set inner Hausdorff loss = 0 if initial estimate is empty
OuterHausdorff(setdiff(1:S,NEmpty_KS)) = Inf;   % set outer Hausdorff loss = Inf if initial estimate is empty

% Check coverage of KS's CS ) %
temp1  = max(repmat(nuP0,1,S)-SupportFunction-repmat(C1',m,1)./sqrt(n));    % directed Hausdorff distance
coverKS = sum(temp1<=0)/S;                                                  % coverage probability of KS's CS
temp2  = max(repmat(nuP0,1,SKS)-SupportFunction(:,NEmpty_KS)-repmat(C1(NEmpty_KS)',m,1)./sqrt(n));
coverKS_ne = sum(temp2<=0)/SKS;                                             % coverage probability of KS's CS cond. on initial estimate being nonempty

% Check volume of KS CS %
AVolKS    = median(VolKS);									% compute median volume of KS CS
AVolKS_ne = median(VolKS(NEmpty_KS));						% compute median volume of KS CS cond. on initial estimate being nonempty

% Check expected Hausdorff distance loss for KS's CS %
Hausdorff_KS = zeros(S,1);
InnerHausdorff_KS = zeros(S,1);
OuterHausdorff_KS = zeros(S,1);
Hausdorff_KS(NEmpty_KS) = max(abs(SupportFunction(:,NEmpty_KS)+repmat(C1(NEmpty_KS)',m,1)./sqrt(n)-repmat(nuP0,1,SKS)));        % compute Hausdorff distance for CS
InnerHausdorff_KS(NEmpty_KS) = max(max(SupportFunction(:,NEmpty_KS)+repmat(C1(NEmpty_KS)',m,1)./sqrt(n)-repmat(nuP0,1,SKS),0)); % compute inner Hausdorff distance for CS
OuterHausdorff_KS(NEmpty_KS) = max(max(repmat(nuP0,1,SKS)-SupportFunction(:,NEmpty_KS)-repmat(C1(NEmpty_KS)',m,1)./sqrt(n),0)); % compute outer Hausdorff distance for CS
Hausdorff_KS(setdiff(1:S,NEmpty_KS)) = Inf;       % set Hausdorff loss = Inf if initial estimate is empty
InnerHausdorff_KS(setdiff(1:S,NEmpty_KS)) = 0;    % set inner Hausdorff loss = 0 if initial estimate is empty
OuterHausdorff_KS(setdiff(1:S,NEmpty_KS)) = Inf;  % set outer Hausdorff loss = Inf if initial estimate is empty

% Average time needed to compute critical value %
At_crit = median(t_prim(NEmpty_KS)+t_dual(NEmpty_KS)+t_eval(NEmpty_KS)+t_crit(NEmpty_KS)+t_weig(NEmpty_KS)); % sum up time needed to calculate estimator and critical value


% Summary of KS estimate %
disp('--- KS Results: check whether correction took place ---')
if SKS==S
    disp('Empty initial estimates not found.')
else
    disp('Empty initial estimates found. ')
    disp('Results conditional on initial estimate being nonempty will be also reported.')
end

% Display results of KS w/o correction %
disp('--- Results on KS w/o correction ---')
disp(['Hausdorff Loss of KS Est                        = ' num2str(median(Hausdorff))])
disp(['Inner Hausdorff Loss of KS Est                  = ' num2str(median(InnerHausdorff))])
disp(['Outer Hausdorff Loss of KS Est                  = ' num2str(median(OuterHausdorff))])
disp(['Coverage Probability of KS CS                   = ' num2str(coverKS)])
disp(['Volume of KS CS                                 = ' num2str(AVolKS)])
disp(['Hausdorff Loss of KS CS                         = ' num2str(median(Hausdorff_KS))])
disp(['Inner Hausdorff Loss of KS CS                   = ' num2str(median(InnerHausdorff_KS))])
disp(['Outer Hausdorff Loss of KS CS                   = ' num2str(median(OuterHausdorff_KS))])
disp(['Size & power of KS test (flat face)             = '])
disp(num2str(spKS_flat))
disp(['Size & power of KS test (kink)                  = '])
disp(num2str(spKS_kink))
disp(['# of nonempty KS initial estimate               = ' num2str(SKS)])
disp(['Average time spent to compute KS critical value = ' num2str(At_crit)])

if SKS < S
% Display results of KS conditonal on initial estimate being nonempty %
disp('--- Results on KS conditional on initial estimate being nonempty ---')
disp(['Hausdorff Loss of KS cond. on nonempty          = ' num2str(median(Hausdorff(NEmpty_KS)))])
disp(['Inner Hausdorff Loss of KS cond. on nonempty    = ' num2str(median(InnerHausdorff(NEmpty_KS)))])
disp(['Outer Hausdorff Loss of KS cond. on nonempty    = ' num2str(median(OuterHausdorff(NEmpty_KS)))])
disp(['Coverage Probability of KS CS cond. on nonempty = ' num2str(coverKS_ne)])
disp(['Volume of KS CS cond. on nonempty               = ' num2str(AVolKS_ne)])
disp(['Hausdorff Loss of KS CS cond. on nonempty       = ' num2str(median(Hausdorff_KS(NEmpty_KS)))])
disp(['Inner Hausdorff Loss of KS CS cond. on nonempty = ' num2str(median(InnerHausdorff_KS(NEmpty_KS)))])
disp(['Outer Hausdorff Loss of KS CS cond. on nonempty = ' num2str(median(OuterHausdorff_KS(NEmpty_KS)))])  
disp(['Size & power of KS test (flat face, nonempty)   = '])
disp(num2str(spKS_flat_ne))
disp(['Size & power of KS test (kink, nonempty)        = '])
disp(num2str(spKS_kink_ne))
end

% Save results %
save(['./mat/MC_n' num2str(n) 'K' num2str(K) 'c' num2str(c) 'S' num2str(S) 'KS.mat'],...
'n','S','SI','B','m','K','c','suppz','tau','A','beta','dt','dF','dtau','dloc',...
'Support*','LM*','St*','C*','Vol*','t_*','NE*','SKS*',...
'cover*','AVol*','Hausdorff*','Inner*','Outer*','sp*','At*');

matlabpool close

