% This program produces Monte Carlo results for uniformly valid procedures
% by Andrews & Soares (2010) and Bugni (2010) applied to Example 1.
% This program requires two free Matlab toolboxes: YALMIP and MPT.
% Please set path to the directories where YALMIP and MPT are saved.
% This code uses Matlab's parallel processing toolbox.
% If multiple processors are not used, comment out line 19 and
% replace "parfor" in line 87 with "for".
% The user must set K (# of support points), c (strength of identification)
% and SI (# of inner loops).
% The program conducts S=5000 Monte Carlo simulations, which are split 
% into SI inner loops to save memory used by YALMIP.
% Pick SI so that S/SI becomes an integer, ideally a multiple of the number
% of processors.

clear
close all
clc
yalmip('clear') % clear memory used by yalmip
matlabpool(4)   % set # of processors for parallel processing

% Set simulations parameters %
n = 500; 	% sample size
S = 5000;   % # of simulation draws
SI= 40;  	% # of inner loops (make sure S/SI is an integer)
B = 200; 	% # of booststrap samples
m = 100;  	% # of grid points on unit sphere
K = 5;   	% # of support points
c = 1;  	% strength of identification
suppz = linspace(-5,5,K); %z's support
tau   = [0, log(log(n)), log(n), n^(1/8), n^(1/4)]; % grid points for slackness
A = [-ones(K,1) -suppz'; ones(K,1) suppz']; % slope of affine constraints

% Set parameter values (DGP)
beta  = [1; 2]; % true parameter       
dt = size(beta,1);	% dimension of parameter
dF = 2*K; 			% number of inequalities
dtau = length(tau); % number of values of tau
dloc = 5;			% number of local alternatives

% Initialize cells & matrices that store results %
SupportFunctionBugni_init = zeros(m,S,dtau);     % a matrix that stores support function of Bugni's initial estimate 
SupportFunctionBugni = zeros(m,S,dtau);          % a matrix that stores support function of Bugni's CI 
StBugni_kink = zeros(S,dloc);                    % a matrix that stores Bugni test stat for a kink point
StBugni_flat = zeros(S,dloc);                    % a matrix that stores Bugni test stat for a point on a flat face
CBugni = zeros(S,dtau);                          % a vector that stores Bugni's critical value for CS 
CBugni_kink = zeros(S,dtau,dloc);                % a vector that stores Andrews & Soares (AS) critical values for a kink point
CBugni_flat = zeros(S,dtau,dloc);                % a vector that stores Andrews & Soares (AS) critical values for a point on a flat face
VolBugni = zeros(S,dtau);                        % a vector that stores Bugni's CI's volume 

% Stopwatch %
t_est_bugni  = zeros(S,dtau);   % time needed to compute initial estimate 
t_crit_bugni = zeros(S,dtau);   % time needed to compute bootstrap critical value for CS 
t_kink_bugni = zeros(S,dtau);   % time needed to compute bootstrap critical value for kink 
t_flat_bugni = zeros(S,dtau);   % time needed to compute bootstrap critical value for flat face 
% Temporary vectors for inner loops %
subSupportFunctionBugni_init = zeros(m,S/SI,dtau); 
subSupportFunctionBugni = zeros(m,S/SI,dtau); 
subStBugni_kink = zeros(S/SI,dloc);
subStBugni_flat = zeros(S/SI,dloc);
subCBugni = zeros(S/SI,dtau);
subCBugni_kink = zeros(S/SI,dtau,dloc);
subCBugni_flat = zeros(S/SI,dtau,dloc);
subvolBugni = zeros(S/SI,dtau);
subt_est_bugni  = zeros(S/SI,dtau);
subt_crit_bugni = zeros(S/SI,dtau);
subt_kink_bugni = zeros(S/SI,dtau);
subt_flat_bugni = zeros(S/SI,dtau);

% Grid the unit sphere %
th = linspace(-pi,pi,m); % angle
q = [sin(th);cos(th)]';  % unit sphere

% Load Population Data %
load(['./mat/population_K' num2str(K) 'c' num2str(c)  '.mat'])
loc_kink = repmat(theta_kink',1,dloc) + repmat([0;1],1,dloc) .* repmat(linspace(0,10,dloc)./sqrt(n),2,1); % kink point and alternatives
loc_flat = repmat(theta_flat',1,dloc) + repmat([1;0],1,dloc) .* repmat(linspace(0,10,dloc)./sqrt(n),2,1); % point on flat face and alternatives

ops  = sdpsettings('cachesolvers',1);
% The beginning of Monte Carlo %

for h=1:SI
    % Define variables used by YALMIP outside the inner loop %
    thetas = sdpvar(dt,1); 	% control variable in the primal problem
    w = sdpvar(dF,1); 		% control variable in the dual problem
    p = sdpvar(dt,1); 		% unit vector
    % The beginning of inner loop using parallel processing
    parfor s=1:S/SI
        si = s+(h-1)*(S/SI);
        stream = RandStream.create('mlfg6331_64','NumStreams',S,'StreamIndices',1:S,'seed',0,'CellOutput',1); 
        RandStream.setDefaultStream(stream{si});
        % Data Generation %
        Z = [ones(n,1) randsample(stream{si},suppz,n,true)']; 	 % generate Z
        e = randn(stream{si},n,1);     							 % generate error for Y
        v = 0.2*rand(stream{si},n,1); 							 % generate random coefficients
        ticweig = tic;
        W = reshape(randsample(stream{si},[-1,1],n*B,true),n,B); % generate bootstrap weights
        subt_weig(s) = toc(ticweig);
        
        % Generate YLi, YUi
        Y  = Z*beta + e;          % generate Y
        YL = Y -v.*Z(:,2).^2 -c ; % generate Y_L
        YU = Y +v.*Z(:,2).^2 +c ; % generate Y_U
        
        % Instrument (indiators) %
        Iz = repmat(Z(:,2),1,length(suppz)) == repmat(suppz,n,1);    	% indicator for each value of z                 
        EYLz = Iz'*YL./n;   											% mean of Y_L * Iz							
        EYUz = Iz'*YU./n;   											% mean of Y_U * Iz
        EIz = mean(Iz)';    											% mean of Iz
        
        % Variance of moments %
        sYLz = mean(((Iz.*repmat(YL,1,K))./(repmat(EIz',n,1))-(Iz./repmat(EIz',n,1)).*repmat((EYLz./EIz)',n,1)).^2)';
        sYUz = mean(((Iz.*repmat(YU,1,K))./(repmat(EIz',n,1))-(Iz./repmat(EIz',n,1)).*repmat((EYUz./EIz)',n,1)).^2)';                              
        
        % Elements of Constraints %
        FL = [-EYLz./EIz; EYUz./EIz];                    % intercept of affine constraints
                
        % Find initial estimator through interpolation of solution to primal problem
        Ngrid = 100;                    % The grid across endpoints
        Theta_hat = cell(dtau,1);       % cell of matrices for Bugni's initial estimator under different values of tau
        theta_p = NaN(m,dt,dtau);       % 3D matrix of boundary points
        t_init_bugni = zeros(1,dtau);   % time for computing initial estimate
        mu_bugni_init = zeros(m,dtau);  % We'll store support function of inital estimator
                        
        for l = 1:dtau
            tic
            % Set up the optimization problem for this tau value
            Fp   = [A*thetas - FL <=  tau(l)*sqrt([sYLz;sYUz])/sqrt(n), -1<=p<=1]; % constraint in primal form
            objp = -p'*thetas;                                                     % objective function in primal form
        
            % Create a solved the optimization problem
            [~, ~, ~, Valuefunction_p, Optimizer_p] ...      % solve the primal problem
            = solvemp(Fp,objp,ops,p);
        
            % Find and Store the boundary points for this choice of tau
            for k = 1:m
                assign(p,q(k,:)');                           % set direction
                temp = double(Optimizer_p);                  % store boundary point 
                if ~isempty(temp)
                    theta_p(k,:,l) = temp';
                    mu_bugni_init(k,l) = -double(Valuefunction_p);
                end
            end
            toc
            
            % Make the Theta hat matrix by interpolating boundary points -
            if ~isnan(sum(theta_p(:,1,l)))                
                % If nonempty call a function that fills vertices at Ngrid
                Theta_hat{l} = ReturnFilled(unique(theta_p(:,:,l),'rows'),Ngrid);                        
            end
                                                                                                                        
            % And store the time. Note otherwise, this stays empty
            t_init_bugni(l) = toc;									% store time spent to construct initial estimate
        end                                                           
        subt_est_bugni(s,:) = t_init_bugni; % store time used to compute estimate                                                                                     
        subSupportFunctionBugni_init(:,s,:) = mu_bugni_init;

        % Compute AS statistic under null and alternative %
		subStBugni_kink(s,:) = sqrt(n)*max(max((A*loc_kink-repmat(FL,1,dloc))./repmat([sYLz.^(1/2); sYUz.^(1/2)],1,dloc),0)); % AS test statistic for kink
		subStBugni_flat(s,:) = sqrt(n)*max(max((A*loc_flat-repmat(FL,1,dloc))./repmat([sYLz.^(1/2); sYUz.^(1/2)],1,dloc),0)); % AS test statistic for flat face
        
        % Generate Bootstrap Samples %
        TBugni = zeros(B,dtau);                 % a vector for bootstrapped statistic 
        TBugni_kink = zeros(B,dtau,dloc);       % a vector for bootstrapped statistic for kink point 
        TBugni_flat = zeros(B,dtau,dloc);       % a vector for bootstrapped statistic for flat face
        t_boot_bugni = zeros(B,1);              % time needed to draw bootstrap sample
        t_stat_bugni = zeros(B,dtau);    		% time needed to compute statistics under each value of tau
		t_kink_bugni = zeros(B,dtau);	     	% time needed to compute statistics on kink under each value of tau
		t_flat_bugni = zeros(B,dtau);	     	% time needed to compute statistics on flat face under each value of tau
                
        % And go through the bootstrap               
        for bb = 1:B
            tic
            ind = randsample(stream{si},1:n,n,true);                    % Generate indices w/ replacement
            YLb = YL(ind);                                              % Bootstrap YL
            YUb = YU(ind);                                              % Bootstrap YU
            Zb  = Z(ind,:);                                             % Bootstrap Z
            Izb = repmat(Zb(:,2),1,length(suppz)) == repmat(suppz,n,1); % indicator for each value of z                                              
            EYLzb = Izb'*YLb./n;                                        % mean of Y_L * Iz
            EYUzb = Izb'*YUb./n;                                        % mean of Y_U * Iz
            EIzb = mean(Izb)';                                          % mean of Iz
            sYLzb = mean(((Izb.*repmat(YLb,1,K))./(repmat(EIzb',n,1))-(Izb./repmat(EIzb',n,1)).*repmat((EYLzb./EIzb)',n,1)).^2)';
            sYUzb = mean(((Izb.*repmat(YUb,1,K))./(repmat(EIzb',n,1))-(Izb./repmat(EIzb',n,1)).*repmat((EYUzb./EIzb)',n,1)).^2)';
            FLb = [-EYLzb./EIzb; EYUzb./EIzb];
            mom = FLb./[sYLzb.^(1/2); sYUzb.^(1/2)] - FL./[sYLz.^(1/2); sYUz.^(1/2)]; %  truncated studentized moments
            t_boot_bugni(bb) = toc;
                                    
            for l=1:dtau
                ticstat = tic;
                if ~isempty(Theta_hat{l})           % if the initial estimate is nonempty
                    % Compute critical value for CS %
                    Np = size(Theta_hat{l},1);      % store the number of points in initial estimate
                    GMS = tau(l)/sqrt(n) >= abs(A*Theta_hat{l}' - repmat(FL,1,Np))./repmat([sYLz.^(1/2); sYUz.^(1/2)],1,Np);  % Indicator for moments on
                    momval = sqrt(n)*(A*Theta_hat{l}'.*repmat(1./[sYLzb.^(1/2); sYUzb.^(1/2)]-1./[sYLz.^(1/2); sYUz.^(1/2)],1,Np) ... 
                             - repmat(mom,1,Np));                                                                             % The moments evaluated (note, max of zero not taken)  
                    Qvalue = max(momval.*GMS)';                                                                               % The function evaluated on Theta hat
                    
                    % Finally compute the Bugni statistic
                    TBugni(bb,l) = max([Qvalue;0]);     % max over ID set to compute the statistic	-- A zero added in case all negative
                    t_stat_bugni(bb,l) = toc(ticstat);  % store how long it took to compute the statistic
                else
                    TBugni(bb,l) = 0;                   % if the initial estimate is empty, set statistic to 0
                    t_stat_bugni(bb,l) = toc(ticstat);  % store how long it took to compute the statistic
                end
                tickink = tic;
                
                % Compute critical values for test %
                GMS_kink = -tau(l)/sqrt(n) < (A*loc_kink-repmat(FL,1,dloc))./repmat([sYLz.^(1/2); sYUz.^(1/2)],1,dloc);       % GMS for point on kink
                momkink = sqrt(n)*max((A*loc_kink)...
                    .*(repmat(1./[sYLzb.^(1/2); sYUzb.^(1/2)]-1./[sYLz.^(1/2); sYUz.^(1/2)],1,dloc)) - repmat(mom,1,dloc),0); % values of moments at point on kink
                TBugni_kink(bb,l,:) = max(momkink .* GMS_kink);                                                               % compute test stat. for point on kink
                t_kink_bugni(bb,l) = toc(tickink);
                ticflat = tic;
                GMS_flat = -tau(l)/sqrt(n) < (A*loc_flat-repmat(FL,1,dloc))./repmat([sYLz.^(1/2); sYUz.^(1/2)],1,dloc);       % GMS for point on flat face
                momflat = sqrt(n)*max((A*loc_flat)...
                    .*(repmat(1./[sYLzb.^(1/2); sYUzb.^(1/2)]-1./[sYLz.^(1/2); sYUz.^(1/2)],1,dloc)) - repmat(mom,1,dloc),0); % values of moments at point on flat face
                TBugni_flat(bb,l,:) = max(momflat .* GMS_flat);                                                               % compute test stat. for point on flat face
                t_flat_bugni(bb,l) = toc(ticflat);
            end                        
        end                
        
        tic
        qBugni    = quantile(TBugni,0.95);   	    % Bugni's critical value
		t_quant_bugni = toc;               			% store how long it took to compute the quantile
		tic
        qBugni_kink = quantile(TBugni_kink,0.95);	% AS's critical value for kink 
		t_quant_kink_bugni = toc;
		tic
        qBugni_flat = quantile(TBugni_flat,0.95);	% AS's critical value for flat face
		t_quant_flat_bugni = toc;
        subCBugni(s,:)      = qBugni;           	% store Bugni's critical value
        subCBugni_kink(s,:,:) = qBugni_kink;        % store Bugni's critical value for kink
        subCBugni_flat(s,:,:) = qBugni_flat;        % store Bugni's critical value for flat face
        subt_crit_bugni(s,:) = repmat(sum(t_boot_bugni),1,dtau) + sum(t_stat_bugni) + t_quant_bugni; 	  % compute how long in total it took to compute the critical value
		subt_kink_bugni(s,:) = repmat(sum(t_boot_bugni),1,dtau) + sum(t_kink_bugni) + t_quant_kink_bugni; % compute how long in total it took to compute the critical value
		subt_flat_bugni(s,:) = repmat(sum(t_boot_bugni),1,dtau) + sum(t_flat_bugni) + t_quant_flat_bugni; % compute how long in total it took to compute the critical value
        
        % Calculate Bugni's CS & Compute its volume %
        vol_bugni       = zeros(1,dtau);			% a vector that stores volumes of Bugni's CS
        mu_bugni        = zeros(m,dtau);			% a vector that stores support function of Bugni's CS
        theta_p_CS      = NaN(m,dt,dtau);
                                         
        for l=1:dtau
            % We'll find the CS by solving for vertices. Note: here the max
            % form (rather than sum) of Qn lets us find it easier
            
            % Set up the optimization problem for this tau value
            Fp   = [A*thetas - FL <=  qBugni(l)*sqrt([sYLz;sYUz])/sqrt(n), -1<=p<=1]; % constraint in primal form
            objp = -p'*thetas;                                                        % objective function in primal form
        
            % Solve the optimization problem
            [~, ~, ~, Valuefunction_p, Optimizer_p] ...      % solve the primal problem under largest tau
            = solvemp(Fp,objp,ops,p);
        
            % Store the boundary points for the CS for this choice of tau
            for k = 1:m
                assign(p,q(k,:)');                           % set direction
                temp = double(Optimizer_p);                  % store boundary point 
                if ~isempty(temp)
                    theta_p_CS(k,:,l) = temp';
                    mu_bugni(k,l) = -double(Valuefunction_p);
                end
            end
            
            % Check if the confidence set is empty or not
            if ~isnan(sum(theta_p_CS(:,1,l)))
                %Get the vertices of the Confidence Set
                tempvertex = unique(theta_p_CS(:,:,l),'rows');
            
                % Use the vertices to compute the volume. Note already took
                % care of getting the support function.
                try
   					[~, V_b]  = convhull(tempvertex(:,1),tempvertex(:,2));  % compute volume of CS
                    vol_bugni(l) = V_b;                                     % store volume of CS
				catch exception
                    vol_bugni(l) = 0;                                       % set volume =0 if the set is degenerated
                end                            
            else
                vol_bugni(l) = 0;                                           % if CS is empty, set volume to 0
            end
        end
        
		subvolBugni(s,:)                    = vol_bugni;
        subSupportFunctionBugni(:,s,:)      = mu_bugni;
    end
    
    % Pass results from the inner loop %
    SupportFunctionBugni_init(:,(h-1)*(S/SI)+1:h*(S/SI),:) = subSupportFunctionBugni_init;
    SupportFunctionBugni(:,(h-1)*(S/SI)+1:h*(S/SI),:) = subSupportFunctionBugni;
    StBugni_kink((h-1)*(S/SI)+1:h*(S/SI),:) = subStBugni_kink;
    StBugni_flat((h-1)*(S/SI)+1:h*(S/SI),:) = subStBugni_flat;
    CBugni((h-1)*(S/SI)+1:h*(S/SI),:) = subCBugni;
    CBugni_kink((h-1)*(S/SI)+1:h*(S/SI),:,:) = subCBugni_kink;
    CBugni_flat((h-1)*(S/SI)+1:h*(S/SI),:,:) = subCBugni_flat;
	VolBugni((h-1)*(S/SI)+1:h*(S/SI),:) = subvolBugni;
    t_est_bugni((h-1)*(S/SI)+1:h*(S/SI),:)  = subt_est_bugni;
    t_crit_bugni((h-1)*(S/SI)+1:h*(S/SI),:) = subt_crit_bugni;
    t_kink_bugni((h-1)*(S/SI)+1:h*(S/SI),:) = subt_kink_bugni;
    t_flat_bugni((h-1)*(S/SI)+1:h*(S/SI),:) = subt_flat_bugni;
                
    yalmip('clear') % Clear the memory used by YALMIP
    % Go to the next inner loop %
    disp(['Current sub-block: ' num2str(h) '/' num2str(SI)])
end
% The end of outer loop (Monte Carlo) %

% Check empty set estimates %
NEmpty_Bugni_init     = cell(dtau,1);	% a cell that stores simulations with non-empty Bugni initial estimate 
NEmpty_Bugni     = cell(dtau,1);        % a cell that stores simulations with non-empty Bugni CS
SBugni_init    = zeros(dtau,1);         % # of simulation draws with non-empty Bugni initial estimate 
SBugni    = zeros(dtau,1);              % # of simulation draws with non-empty Bugni CS
for l=1:dtau
   NEmpty_Bugni_init{l} = find(sum(abs(SupportFunctionBugni_init(:,:,l))~=0)); % find simulations with non-empty initial estimate
   NEmpty_Bugni{l}    = find(sum(abs(SupportFunctionBugni(:,:,l))~=0));		   % find simulations with non-empty Bugni CS
   SBugni_init(l) = size(NEmpty_Bugni_init{l},2)';                             % calculate # of simulation draws with non-emptyinitial estiamte
   SBugni(l) = size(NEmpty_Bugni{l},2)';								       % calculate # of simulation draws with non-empty Bugni CS
end

% Compute size and power %
spBugni_kink = zeros(dtau,dloc);
spBugni_flat = zeros(dtau,dloc);
for vv=1:dloc
spBugni_kink(:,vv) = mean(repmat(StBugni_kink(:,vv),1,dtau)-CBugni_kink(:,:,vv)>0)'; % compute size & power when theta is on kink
spBugni_flat(:,vv) = mean(repmat(StBugni_flat(:,vv),1,dtau)-CBugni_flat(:,:,vv)>0)'; % compute size & power when theta is on flat face
end
spBugni_kink = spBugni_kink'; % each row gives size & power for each value of tau (kink)
spBugni_flat = spBugni_flat'; % each row gives size & power for each value of tau (flat face)


% Check median Hausdorff distance loss for Bugni's initial estimates (without correction)%
Hausdorff_Bugni_init = zeros(S,dtau);
InnerHausdorff_Bugni_init = zeros(S,dtau);
OuterHausdorff_Bugni_init = zeros(S,dtau);
MedHausdorff_Bugni_init_ne = zeros(1,dtau);
MedInnerHausdorff_Bugni_init_ne = zeros(1,dtau);
MedOuterHausdorff_Bugni_init_ne = zeros(1,dtau);
for l=1:dtau
Hausdorff_Bugni_init(NEmpty_Bugni_init{l},l) = max(abs(SupportFunctionBugni_init(:,NEmpty_Bugni_init{l},l)-repmat(nuP0,1,SBugni_init(l))));
InnerHausdorff_Bugni_init(NEmpty_Bugni_init{l},l) = max(max(SupportFunctionBugni_init(:,NEmpty_Bugni_init{l},l)-repmat(nuP0,1,SBugni_init(l)),0));
OuterHausdorff_Bugni_init(NEmpty_Bugni_init{l},l) = max(max(repmat(nuP0,1,SBugni_init(l))-SupportFunctionBugni_init(:,NEmpty_Bugni_init{l},l),0));
Hausdorff_Bugni_init(setdiff(1:S,NEmpty_Bugni_init{l}),l) = Inf;                                % set Hausdorff loss = Inf if initial estimate is empty
InnerHausdorff_Bugni_init(setdiff(1:S,NEmpty_Bugni_init{l}),l) = 0;                             % set inner Hausdorff loss = 0 if initial estimate is empty
OuterHausdorff_Bugni_init(setdiff(1:S,NEmpty_Bugni_init{l}),l) = Inf;                           % set outer Hausdorff loss = Inf if initial estimate is empty
MedHausdorff_Bugni_init_ne(l) = median(Hausdorff_Bugni_init(NEmpty_Bugni_init{l},l));           % compute median hausdorff loss
MedInnerHausdorff_Bugni_init_ne(l) = median(InnerHausdorff_Bugni_init(NEmpty_Bugni_init{l},l)); % compute median inner hausdorff loss
MedOuterHausdorff_Bugni_init_ne(l) = median(OuterHausdorff_Bugni_init(NEmpty_Bugni_init{l},l)); % compute median outer hausdorff loss
end

% Check coverage of Bugni's CS %
coverBugni = zeros(1,dtau);
coverBugni_ne = zeros(1,dtau);
for l=1:dtau
temp3 = max(repmat(nuP0,1,S)-SupportFunctionBugni(:,:,l));                          % directed Hausdorff distance
temp4 = max(repmat(nuP0,1,SBugni(l))-SupportFunctionBugni(:,NEmpty_Bugni{l},l));    % directed Hausdorff distance when initial estimate is nonempty
coverBugni(l) = sum(temp3<=0)/S;                        % coverage probability of Bugni's CS
coverBugni_ne(l) = sum(temp4<=0)/SBugni(l);             % coverage probabiligy of Bugni's CS cond. on initial estimate being nonempty
end

% Check volume of Bugni CS %
AVolBugni = zeros(1,dtau);
AVolBugni_ne = zeros(1,dtau);
for l=1:dtau
AVolBugni(l) = median(VolBugni(:,l));					% compute median volume of Bugni CS
AVolBugni_ne(l) = median(VolBugni(NEmpty_Bugni{l},l));	% compute median volume of Bugni CS cond. on initial estimate being nonempty
end

% Check median Hausdorff distance loss for Bugni's CS %
Hausdorff_Bugni = zeros(S,dtau);
InnerHausdorff_Bugni = zeros(S,dtau);
OuterHausdorff_Bugni = zeros(S,dtau);
MedHausdorff_Bugni_ne = zeros(1,dtau);
MedInnerHausdorff_Bugni_ne = zeros(1,dtau);
MedOuterHausdorff_Bugni_ne = zeros(1,dtau);
for l=1:dtau
Hausdorff_Bugni(NEmpty_Bugni{l},l) = max(abs(SupportFunctionBugni(:,NEmpty_Bugni{l},l)-repmat(nuP0,1,SBugni(l))));
InnerHausdorff_Bugni(NEmpty_Bugni{l},l) = max(max(SupportFunctionBugni(:,NEmpty_Bugni{l},l)-repmat(nuP0,1,SBugni(l)),0));
OuterHausdorff_Bugni(NEmpty_Bugni{l},l) = max(max(repmat(nuP0,1,SBugni(l))-SupportFunctionBugni(:,NEmpty_Bugni{l},l),0));
MedHausdorff_Bugni_ne(l) = median(Hausdorff_Bugni(NEmpty_Bugni{l},l));           % compute median hausdorff loss cond. on initial estimate being nonempty
MedInnerHausdorff_Bugni_ne(l) = median(InnerHausdorff_Bugni(NEmpty_Bugni{l},l)); % compute median inner hausdorff loss cond. on initial estimate being nonempty
MedOuterHausdorff_Bugni_ne(l) = median(OuterHausdorff_Bugni(NEmpty_Bugni{l},l)); % compute median outer hausdorff loss cond. on initial estimate being nonempty
end

% Report median time needed to construct CS %
At_crit_bugni = zeros(1,dtau);
for l=1:dtau
At_crit_bugni(l) = median(t_est_bugni(NEmpty_Bugni_init{l},l) + t_crit_bugni(NEmpty_Bugni_init{l},l)); % median of time needed to compute estimate & critical value
end


% Summary of Bugni estimate %
disp('--- Bugni Results: check whether correction took place ---')
if min(SBugni_init == repmat(S,dtau,1))
    disp('Empty initial estimates not found.')
else
    disp('Empty initial estimates found. ')
    disp('Results conditional on initial estimate being nonempty will be also reported.')
end
% Display results on Bugni w/o correction %
disp('--- Results on Bugni w/o correction ---')
disp(['Hausdorff Loss of Bugni estimate                = ' num2str(median(Hausdorff_Bugni_init))])
disp(['Inner Hausdorff Loss of Bugni estimate          = ' num2str(median(InnerHausdorff_Bugni_init))])
disp(['Outer Hausdorff Loss of Bugni estimate          = ' num2str(median(OuterHausdorff_Bugni_init))])
disp(['Coverage Probability of Bugni CS                = ' num2str(coverBugni)])
disp(['Volume of Bugni CS                              = ' num2str(AVolBugni)])
disp(['Hausdorff Loss of Bugni CS                      = ' num2str(median(Hausdorff_Bugni))])
disp(['Inner Hausdorff Loss of Bugni CS                = ' num2str(median(InnerHausdorff_Bugni))])
disp(['Outer Hausdorff Loss of Bugni CS                = ' num2str(median(OuterHausdorff_Bugni))])
disp(['Size & power of AS test (flat face)             = '])
disp(num2str(spBugni_flat))
disp(['Size & power of AS test (kink)                  = '])
disp(spBugni_kink)
disp(['Average time spent to compute Bugni cvs         = ' num2str(At_crit_bugni)])
disp(['# of nonempty Bugni initial estimate            = ' num2str(SBugni_init')])

if max(SBugni_init == repmat(S,dtau,1))
% Display results on Bugni conditional on initial estimate being nonempty %
disp('--- Results on Bugni conditional on initial estimate being nonempty ---')
disp(['Hausdorff Loss of Bugni estimate cond. on ne    = ' num2str(MedHausdorff_Bugni_init_ne)])
disp(['Inner Hausdorff Loss of Bugni estimate (ne)     = ' num2str(MedInnerHausdorff_Bugni_init_ne)])
disp(['Outer Hausdorff Loss of Bugni estimate (ne)     = ' num2str(MedOuterHausdorff_Bugni_init_ne)])
disp(['Coverage Probability of Bugni CS cond. nonempty = ' num2str(coverBugni_ne)])
disp(['Volume of Bugni CS w/o correction               = ' num2str(AVolBugni_ne)])
disp(['Hausdorff Loss of Bugni CS cond. on ne          = ' num2str(MedHausdorff_Bugni_ne)])
disp(['Inner Hausdorff Loss of Bugni CS cond. on ne    = ' num2str(MedInnerHausdorff_Bugni_ne)])
disp(['Outer Hausdorff Loss of Bugni CS cond. on ne    = ' num2str(MedOuterHausdorff_Bugni_ne)])
end


% Save results %
save(['./mat/MC_n' num2str(n) 'K' num2str(K) 'c' num2str(c) 'S' num2str(S) 'Bugni.mat'],...
'n','S','SI','B','m','K','c','suppz','tau','A','beta','dt','dF','dtau','dloc',...
'Support*','St*','C*','Vol*','t_*','NE*','SBugni*',...
'cover*','AVol*','Hausdorff*','Inner*','Outer*','sp*','At*','Med*');

matlabpool close

