% This program computes the population support function (nuP0),
% Lagrange multiplier (lmP0), the two boundary points of the identified set
% used for hypotheisis testing (theta_flat and theta_kink).
% These objects will be saved in population_K<value of K>c<value of c>.mat. 
% This program requires two free Matlab toolboxes: YALMIP and MPT.

clear
close all
clc

% Set parameter values (DGP)
c = 1;  % strength of identification
K = 5;  % # of support points 
suppz = linspace(-5,5,K); %Z's support
beta  = [1; 2]; % true parameter
Sigma = [1 0    % covariance of (epsL,epsU)
    0 1];
EyLz = ([ones(K,1),suppz']*beta-c-0.1*(suppz.^2)')...
    .*(ones(K,1)./K); % expected value of Y_L times indicator 
EyUz = ([ones(K,1),suppz']*beta+c+0.1*(suppz.^2)')...
    .*(ones(K,1)./K); % expected value of Y_U times indicator 
E1z = ones(K,1)./K;   % expected value of the indicator
disp('--- Model Information ---')
disp(['# of parameters (d_theta):' num2str(size(beta,1))])
disp(['# of constraints (d_F):' num2str(2*K)])
disp(['# of moments (d_m):' num2str(3*K)])

% Set constraints %
A = [-ones(K,1) -suppz'; ones(K,1) suppz']; % slope of affine constraints
FL = [-EyLz./E1z;EyUz./E1z];                % intercept of affine constraints
[dF dt]=size(A);

% Solve the dual problem %
disp('---Solve the dual problem---')
tic
w = sdpvar(dF,1); % control (dual) variable 
p = sdpvar(dt,1); % parameter
F = [A'*w == p, w>=0,-1<=p<=1]; % constraint
obj = w'*FL;      % objective function
sdpsettings('solver','gplk');
[sol,diagnostics,aux,Valuefunction,Optimizer]...
    = solvemp(F,obj,[],p); % solve the dual problem
toc

% Grid the unit sphere %
m = 100;                  % # of grid points
th = linspace(-pi,pi,m);  % angle
q = [sin(th);cos(th)]';   % 2-dimensional sphere

% Evaluate the support function and Lagrange multipliers %
disp(['---Evaluate the Support Function & LM on ' num2str(m) ' grid points---'])
lambda = zeros(m,dF); % a matrix that stores Lagrange multiplier
mu = zeros(m,1);      % a vector that stores support function
tic
for k = 1:m
assign(p,q(k,:)');            % set direction
temp = double(Optimizer);     % store Lagrange multiplier
lambda(k,:) = temp';          
mu(k) = double(Valuefunction);% store support function
end
toc

% Obtain points for hypothesis test %
theta = sdpvar(dt,1); 
Fp   = [A*theta - FL <=  0, -1<=p<=1];     % constraint in primal form
objp = -p'*theta;                          % objective function in primal form
[sol_p, diagnostics_p, aux_p, Valuefunction_p, Optimizer_p] ...  % solve the primal problem
    = solvemp(Fp,objp,[],p);
assign(p,[0;1]);                           % set direction
temp = double(Optimizer_p);                % store boundary point
theta_kink = temp';
assign(p,[1;0]);                           % set direction
temp = -double(Valuefunction_p);    
theta_flat = [temp, 2];


% Save Support Function & LM%
nuP0 = mu;      % population support function
lmP0 = lambda;  % population Lagrange multiplier
save(['./mat/population_K' num2str(K) 'c' num2str(c)  '.mat'],...
    'nuP0', 'lmP0', 'theta_flat', 'theta_kink')

% Diagnostics %
disp('--- Diagnostics ---')
disp(diagnostics.info)